""" Module for generating reports. """
# clean
import copy
import time
import os
from typing import Any, Optional, List, Union
from reportlab.lib.enums import TA_JUSTIFY, TA_CENTER
from reportlab.lib.pagesizes import letter
from reportlab.platypus import Paragraph, Spacer, Image, PageBreak, Table
from reportlab.platypus.doctemplate import PageTemplate, BaseDocTemplate
from reportlab.platypus.frames import Frame
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.units import inch, cm, mm
from reportlab.platypus.tableofcontents import TableOfContents
from hisim import utils


class MyDocTemplate(BaseDocTemplate):

    """MyDocTemplate class."""

    def __init__(self, filename, **kw):
        """Initialize the doc template."""
        self.allow_splitting = 0
        super().__init__(filename, **kw)
        self.template = PageTemplate("normal", [Frame(2.5 * cm, 2.5 * cm, 15 * cm, 25 * cm, id="F1")])
        self.addPageTemplates(self.template)

    # Entries to the table of contents can be done either manually by
    # calling the addEntry method on the TableOfContents object or automatically
    # by sending a 'TOCEntry' notification in the afterFlowable method of
    # the DocTemplate you are using. The data to be passed to notify is a list
    # of three or four items countaining a level number, the entry text, the page
    # number and an optional destination key which the entry should point to.
    # This list will usually be created in a document template's method like
    # afterFlowable(), making notification calls using the notify() method
    # with appropriate data.

    def afterFlowable(self, flowable):
        """Registers TOC entries."""
        if flowable.__class__.__name__ == "Paragraph":
            text = flowable.getPlainText()
            style = flowable.style.name
            if style == "Heading1":
                self.notify("TOCEntry", (0, text, self.page))
            if style == "Heading2":
                self.notify("TOCEntry", (1, text, self.page))


class ReportGenerator:

    """Class for generating reports."""

    def __init__(self, dirpath: str) -> None:
        """Initialize the pdf report."""
        if dirpath is None:
            raise ValueError("Result path for the report was none.")
        self.story: Any
        self.toc = TableOfContents()

        self.filepath = os.path.join(dirpath, "report.pdf")
        self.open()
        self.write_preamble()
        self.write_table_of_content()
        self.close()

    def open(self):
        """Open a file."""
        self.doc = MyDocTemplate(
            self.filepath, pagesize=letter, rightMargin=72, leftMargin=72, topMargin=72, bottomMargin=18,
        )

        self.styles = getSampleStyleSheet()
        self.styles.add(ParagraphStyle(name="Justify", alignment=TA_JUSTIFY))
        self.styles.add(ParagraphStyle(name="Normal_CENTER", parent=self.styles["Normal"], alignment=TA_CENTER))
        self.styles.add(
            ParagraphStyle(
                name="toc_centered",
                parent=self.styles["Normal"],
                fontSize=20,
                leading=16,
                alignment=TA_CENTER,
                spaceAfter=40,
            )
        )
        self.style_h1 = ParagraphStyle(name="Heading1", fontSize=12, leading=16, spaceBefore=20)
        self.style_h2 = ParagraphStyle(name="Heading2", fontSize=12, leading=14, spaceBefore=10)

    def write_table_of_content(self):
        """Write Table of Content."""

        # Create an instance of TableOfContents. Override the level styles (optional)
        # and add the object to the story

        self.toc = TableOfContents()

        self.toc.levelStyles = [self.style_h1, self.style_h2]

        self.story.append(Paragraph("<b>Table of contents</b>", self.styles["toc_centered"]))
        self.story.append(self.toc)
        self.story.append(PageBreak())

    def write_preamble(self):
        """Write the preamble."""
        # Configuration taken mostly from following tutorial
        # https://www.blog.pythonlibrary.org/2010/03/08/a-simple-step-by-step-reportlab-tutorial/
        story = []

        # Inserts HiSim logo
        logo = os.path.join(utils.hisim_postprocessing_img, "hisim_logo.png")
        im1 = Image(logo, 2 * inch, inch)
        im1.hAlign = "LEFT"

        # Inserts FZJ logo
        logo = os.path.join(utils.hisim_postprocessing_img, "fzj_logo.jpg")
        im2 = Image(logo, 2 * inch, inch)
        im2.hAlign = "RIGHT"

        data = [[im1, im2]]
        report_table = Table(data)
        story.append(report_table)
        story.append(Spacer(1, 50))

        # Insert Title
        titel = "HiSim Simulation Report"
        ptext = f'<font size="30">{titel}</font>'
        story.append(Paragraph(ptext, self.styles["Title"]))
        story.append(Spacer(1, 150))

        # Inserts authors
        authors = [
            "Developers:",
            "\n",
            "Dr. Noah Pflugradt",
            "Dr. Sebastian Dickler",
            "Kevin Knosala",
            "Katharina Rieck",
            "Johanna Ganglbauer",
            "David Neuroth",
            "Tjarko Tjaden",
            "Vitor Hugo Bellotto Zago",
            "Maximilian Hillen",
            "Frank Burkrad",
            "Marwa Alfouly",
            "Franz Oldopp",
            "Markus Blasberg",
        ]

        for part in authors:
            ptext = f'<font size="16">{part.strip()}</font>'
            story.append(Paragraph(ptext, self.styles["Normal"]))
            story.append(Spacer(1, 10))
        story.append(Spacer(1, 30))

        # Inserts address
        address_parts = [
            "Forschungszentrum Jülich",
            "Institute of Energy and Climate Research",
            "Techno - Economic Systems Analysis (IEK - 3)",
            "Wilhelm - Johnen - Straße",
            "52428 Jülich",
            "Germany",
        ]
        for part in address_parts:
            ptext = f'<font size="16">{part.strip()}</font>'
            story.append(Paragraph(ptext, self.styles["Normal"]))
            story.append(Spacer(1, 10))
        story.append(Spacer(1, 30))

        # Inserts time
        formatted_time = time.ctime()
        ptext = f'<font size="16">{formatted_time}</font>'
        story.append(Paragraph(ptext, self.styles["Normal"]))
        story.append(Spacer(1, 30))

        if hasattr(self, "executation_time"):
            # formatted_time
            ptext = f'<font size="16">{formatted_time}</font>'
            story.append(Paragraph(ptext, self.styles["Normal"]))
            story.append(Spacer(1, 30))
        self.story = story
        self.story.append(PageBreak())

    def get_story(self):
        """Get the story."""
        self.story = copy.deepcopy(self.story)

    def write_with_normal_alignment(self, text: Union[List[str], List[Optional[str]]]) -> None:
        """Write a paragraph."""
        if len(text) != 0:
            for part in text:
                if part is not None:
                    if not isinstance(part, str):
                        raise ValueError("Got a non-string somehow: " + str(part))
                    ptext = f'<font size="12">{part.strip()}</font>'
                    self.story.append(Paragraph(ptext, self.styles["Normal"]))
                else:
                    raise ValueError("text contains Nones. Text was: " + str(text))
            self.story.append(Spacer(1, 10))
        self.story.append(Spacer(1, 20))

    def write_with_center_alignment(self, text: List[str]) -> None:
        """Write a paragraph."""
        if len(text) != 0:
            for part in text:
                ptext = f'<font size="12">{part.strip()}</font>'
                paragraph = Paragraph(ptext, self.styles["Normal_CENTER"])
                self.story.append(paragraph)
            self.story.append(Spacer(1, 10))

    def write_figures_to_report(self, file_path: str) -> None:
        """Add figure to the report."""

        if os.path.isfile(file_path):
            image = Image(file_path, useDPI=True)
            image.hAlign = "CENTER"
            self.story.append(image)
        else:
            raise ValueError("no files found")

    def write_figures_to_report_with_size_four_six(self, file_path: str) -> None:
        """Add figure to the report with certain size."""

        if os.path.isfile(file_path):
            image = Image(file_path, width=4 * inch, height=6 * inch)
            image.hAlign = "CENTER"
            self.story.append(image)
        else:
            raise ValueError("no files found")

    def write_figures_to_report_with_size_seven_four(self, file_path: str) -> None:
        """Add figure to the report with certain size."""

        if os.path.isfile(file_path):
            image = Image(file_path, width=7 * inch, height=4 * inch)
            image.hAlign = "CENTER"
            self.story.append(image)
        else:
            raise ValueError("no files found")

    def write_tables_to_report(self, table_as_list_of_lists: List) -> None:
        """Add table to the report."""

        table = Table(
            table_as_list_of_lists,
            style=[
                ("LINEABOVE", (0, 0), (-1, 0), 1, "black"),
                ("LINEABOVE", (0, 1), (-1, 1), 0.5, "black"),
                ("LINEBELOW", (0, -1), (-1, -1), 1, "black"),
            ],
        )
        self.story.append(table)

    def write_heading_with_style_heading_one(self, text: List[str]) -> None:
        """Write text as heading."""
        if len(text) != 0:
            for part in text:
                ptext = f"<b>{part.strip()}</b>"
                self.story.append(Paragraph(ptext, self.style_h1))
            self.story.append(Spacer(1, 10))
        self.story.append(Spacer(1, 30))

    def write_heading_with_style_heading_two(self, text: List[str]) -> None:
        """Write text as heading."""
        if len(text) != 0:
            for part in text:
                ptext = f"<b>{part.strip()}</b>"
                self.story.append(Paragraph(ptext, self.style_h2))
            self.story.append(Spacer(1, 10))

    def page_break(self):
        """Make a page break."""
        self.story.append(PageBreak())

    def add_spacer(self):
        """Add spacer."""
        self.story.append(Spacer(1, 30))

    def add_page_number(self, canvas, doc):
        """Add page number to report."""
        canvas.saveState()
        canvas.setFont(self.styles["Heading2"].fontName, self.styles["Heading2"].fontSize)
        page_number_text = f"{doc.page}"
        canvas.drawRightString(200 * mm, 20 * mm, page_number_text)
        canvas.restoreState()

    def close(self):
        """Close the report."""
        story = copy.deepcopy(self.story)
        self.doc.template.onPage = self.add_page_number
        self.doc.template.onPageEnd = self.add_page_number
        self.doc.multiBuild(story)



"""Data Collection for Scenario Comparison."""
# clean
import glob
import datetime
import os
from typing import Dict, Any, Optional, List, Tuple
import json
import shutil
import re
from collections import defaultdict
import pandas as pd
from hisim import log
from hisim.postprocessing.scenario_evaluation.result_data_processing import (
    ResultDataProcessingModeEnum,
    ResultDataTypeEnum,
    DataFormatEnum,
)


class ResultDataCollection:

    """ResultDataCollection class which collects and concatenate the result data from the system_setups/results."""

    def __init__(
        self,
        scenario_analysis_config_name: str,
        data_format_type: str,
        data_processing_mode: Any,
        simulation_duration_to_check: str,
        time_resolution_of_data_set: Any,
        folder_from_which_data_will_be_collected: str = os.path.join(
            os.pardir, os.pardir, os.pardir, "system_setups", "results"
        ),
        path_to_default_config: Optional[str] = None,
    ) -> None:
        """Initialize the class."""
        result_folder = folder_from_which_data_will_be_collected
        self.result_data_folder = os.path.join(
            os.getcwd(), os.pardir, os.pardir, os.pardir, "system_setups", "scenario_comparison", "data",
        )
        if not os.path.exists(self.result_data_folder):
            os.makedirs(self.result_data_folder)

        # in each system_setups/results folder should be one system setup that was executed with the default config
        self.path_of_scenario_data_executed_with_default_config: str = ""
        self.data_format_type: str = data_format_type
        self.scenario_analysis_config_name: str = scenario_analysis_config_name

        log.information(f"Checking results from folder: {result_folder}")

        list_with_result_data_folders = self.get_only_useful_data(result_path=result_folder)

        if data_processing_mode == ResultDataProcessingModeEnum.PROCESS_ALL_DATA.name:
            parameter_key = None

        elif data_processing_mode == ResultDataProcessingModeEnum.PROCESS_FOR_DIFFERENT_BUILDING_CODES.name:
            parameter_key = "building_code"

        else:
            raise ValueError("Analysis mode is not part of the ResultDataProcessingModeEnum class.")

        log.information(f"Data Collection Mode is {data_processing_mode}")

        print("parameter key ", parameter_key)
        print("##################")

        if path_to_default_config is None:
            list_with_parameter_key_values = None
            list_with_csv_files = list_with_result_data_folders
            list_with_module_config_dicts = None

        else:
            # path to default config is given (which means there should be also a module config dict in the json file in the result folder which has read the config)

            default_config_dict = self.get_default_config(path_to_default_config=path_to_default_config)

            (
                list_with_csv_files,
                list_with_parameter_key_values,
                list_with_module_config_dicts,
            ) = self.go_through_all_result_data_folders_and_collect_file_paths_according_to_parameters(
                list_with_result_data_folders=list_with_result_data_folders,
                default_config_dict=default_config_dict,
                parameter_key=parameter_key,
            )

        if not list_with_csv_files:
            raise ValueError("list_with_csv_files is empty")

        all_csv_files = self.import_data_from_file(
            paths_to_check=list_with_csv_files, analyze_yearly_or_hourly_data=time_resolution_of_data_set,
        )

        dict_of_csv_data = self.make_dictionaries_with_simulation_duration_keys(
            simulation_duration_to_check=simulation_duration_to_check, all_csv_files=all_csv_files,
        )

        self.filepath_of_aggregated_dataframe = self.alternative_read_csv_and_generate_pandas_dataframe(
            dict_of_csv_to_read=dict_of_csv_data,
            time_resolution_of_data_set=time_resolution_of_data_set,
            rename_scenario=True,
            parameter_key=parameter_key,
            list_with_parameter_key_values=list_with_parameter_key_values,
            list_with_module_config_dicts=list_with_module_config_dicts,
        )

        print("\n")

    def get_only_useful_data(self, result_path: str) -> List[str]:
        """Go through all result folders and filter only useful data and write unuseful data into txt file."""

        # go through result path and if the dirs do not contain finished.flag ask for deletion
        self.clean_result_directory_from_unfinished_results(result_path=result_path)

        # get result folders with result data folder
        list_with_all_paths_to_check = self.get_list_of_all_relevant_folders_or_files(
            result_path=result_path, folder_or_filename="result_data_for_scenario_evaluation"
        )
        print(
            "len of list with all paths to containing result data ", len(list_with_all_paths_to_check),
        )
        # filter out results that had buildings that were too hot or too cold
        list_with_all_paths_to_check_after_filtering = self.filter_results_that_failed_to_heat_or_cool_building_sufficiently(
            list_of_result_path_that_contain_scenario_data=list_with_all_paths_to_check
        )
        print(
            "len of list with all paths after filtering ", len(list_with_all_paths_to_check),
        )
        # check if duplicates are existing and ask for deletion
        list_with_result_data_folders = self.go_through_all_scenario_data_folders_and_check_if_module_configs_are_double_somewhere(
            list_of_result_folder_paths_to_check=list_with_all_paths_to_check_after_filtering
        )
        print(
            "len of list with all paths after double checking for duplicates ", len(list_with_result_data_folders),
        )
        return list_with_result_data_folders

    def clean_result_directory_from_unfinished_results(self, result_path: str) -> None:
        """When a result folder does not contain the finished_flag, it will be removed from the system_setups/result folder."""
        list_of_unfinished_folders = []
        file_name = os.path.join(self.result_data_folder, "failed_simualtions.txt")
        mode = "a" if os.path.exists(file_name) else "w"
        with open(file_name, mode, encoding="utf-8",) as file:
            file.write(str(datetime.datetime.now()) + "\n")
            file.write("Failed simulations found in the following folders: \n")
            list_with_all_potential_finsihed_flag_files = self.get_list_of_all_relevant_folders_or_files(
                result_path=result_path, folder_or_filename="finished.flag"
            )
            for filename in list_with_all_potential_finsihed_flag_files:
                if not os.path.exists(filename):
                    file.write(os.path.join(filename) + "\n")
                    list_of_unfinished_folders.append(filename)

            file.write(
                f"Total number of failed simulations in path {result_path}: {len(list_of_unfinished_folders)}"
                + "\n"
                + "\n"
            )

        print(
            f"The following result folders do not contain the finished flag: {list_of_unfinished_folders} Number: {len(list_of_unfinished_folders)}. "
        )

        # if list of unfinished folders is not empty
        if list_of_unfinished_folders:
            answer = input("Do you want to delete them?")
            if answer.upper() in ["Y", "YES"]:
                for filename in list_of_unfinished_folders:
                    shutil.rmtree(os.path.join(result_path, filename))
                print("All folders with failed simulations deleted.")
            elif answer.upper() in ["N", "NO"]:
                print("The folders won't be deleted.")
            else:
                print("The answer must be yes or no.")

    def filter_results_that_failed_to_heat_or_cool_building_sufficiently(
        self, list_of_result_path_that_contain_scenario_data: List[str]
    ) -> List[str]:
        """When a result shows too high or too low building temperatures, it will be filtered and removed from further analysis."""
        list_of_unsuccessful_folders = []
        with open(
            os.path.join(
                self.result_data_folder, "succeeded_simulations_that_showed_too_high_or_too_low_building_temps.txt",
            ),
            "a",
            encoding="utf-8",
        ) as file:
            file.write(str(datetime.datetime.now()) + "\n")
            file.write("Simulations with unsuccessful heating or cooling found in the following folders: \n")
            file.write(
                "Building is too...,"
                "set temperature heating [°C],"
                "set temperature cooling [°C],"
                "min building air temperature [°C],"
                "max building air temperature [°C],"
                "temp deviation below set heating [°C*h],"
                "temp deviation above set cooling [°C*h], folder \n"
            )
            for folder in list_of_result_path_that_contain_scenario_data:
                scenario_data_information = os.path.join(folder, "data_information_for_scenario_evaluation.json")
                main_folder = os.path.normpath(folder + os.sep + os.pardir)
                all_kpis_json_file = os.path.join(main_folder, "all_kpis.json")

                # get set temperatures used in the simulation
                if os.path.exists(scenario_data_information):
                    with open(scenario_data_information, "r", encoding="utf-8") as data_info_file:
                        try:
                            kpi_data = json.load(data_info_file)
                        except Exception as exc:
                            content = data_info_file.read()
                            if content.strip() == "":
                                raise ValueError("The json file is empty. Maybe run the simulation again. "
                                      f"The concerned folder is {folder}") from exc
                        component_entries = kpi_data["componentEntries"]
                        for component in component_entries:
                            if "Building" in component["componentName"]:
                                set_heating_temperature = float(
                                    component["configuration"].get("set_heating_temperature_in_celsius")
                                )
                                set_cooling_temperature = float(
                                    component["configuration"].get("set_cooling_temperature_in_celsius")
                                )
                                break
                else:
                    raise FileNotFoundError(f"The file {scenario_data_information} could not be found. ")

                # open the webtool kpis and check if building got too hot or too cold
                if os.path.exists(all_kpis_json_file):
                    with open(all_kpis_json_file, "r", encoding="utf-8") as kpi_file:
                        kpi_data = json.load(kpi_file)
                        # check if min and max temperatures are too low or too high
                        min_temperature = float(
                            kpi_data["Building"]["Minimum building indoor air temperature reached"].get("value")
                        )
                        max_temperature = float(
                            kpi_data["Building"]["Maximum building indoor air temperature reached"].get("value")
                        )
                        temp_deviation_below_set = kpi_data["Building"][
                            "Temperature deviation of building indoor air temperature being below set temperature 20.0 Celsius"
                        ].get("value")
                        temp_deviation_above_set = kpi_data["Building"][
                            "Temperature deviation of building indoor air temperature being above set temperature 25.0 Celsius"
                        ].get("value")
                        if (
                            min_temperature <= set_heating_temperature - 5.0
                            and max_temperature >= set_cooling_temperature + 5.0
                        ):
                            file.write(
                                "too cold and too warm,"
                                f"{set_heating_temperature},"
                                f"{set_cooling_temperature},"
                                f"{min_temperature},"
                                f"{max_temperature},"
                                f"{temp_deviation_below_set},"
                                f"{temp_deviation_above_set}, {folder}" + "\n"
                            )
                            list_of_unsuccessful_folders.append(folder)
                        elif (
                            min_temperature <= set_heating_temperature - 5.0
                            and max_temperature < set_cooling_temperature + 5.0
                        ):
                            file.write(
                                "too cold,"
                                f"{set_heating_temperature},"
                                f"{set_cooling_temperature},"
                                f"{min_temperature},"
                                f"{max_temperature},"
                                f"{temp_deviation_below_set},"
                                f"{temp_deviation_above_set}, {folder}" + "\n"
                            )
                            list_of_unsuccessful_folders.append(folder)
                        elif (
                            min_temperature > set_heating_temperature - 5.0
                            and max_temperature >= set_cooling_temperature + 5.0
                        ):
                            file.write(
                                "too warm,"
                                f"{set_heating_temperature},"
                                f"{set_cooling_temperature},"
                                f"{min_temperature},"
                                f"{max_temperature},"
                                f"{temp_deviation_below_set},"
                                f"{temp_deviation_above_set}, {folder}" + "\n"
                            )
                            list_of_unsuccessful_folders.append(folder)
                else:
                    raise FileNotFoundError(f"The file {all_kpis_json_file} could not be found. ")

            file.write(
                f"Total number of simulations that have building temperatures way below or above set temperatures: {len(list_of_unsuccessful_folders)}"
                + "\n"
                + "\n"
            )

        print(
            f"Total number of simulations that have building temperatures way below or above set temperatures: {len(list_of_unsuccessful_folders)}"
        )

        # ask if these simulation results should be analyzed
        answer = input("Do you want to take these simulation results into account for further analysis?")
        if answer.upper() in ["N", "NO"]:
            for folder in list_of_unsuccessful_folders:
                list_of_result_path_that_contain_scenario_data.remove(folder)
            print(
                "The folders with too low or too high building temperatures will be discarded from the further analysis."
            )
        elif answer.upper() in ["Y", "YES"]:
            print("The folders with too low or too high building temperatures will be kept for the further analysis.")
        else:
            print("The answer must be yes or no.")

        return list_of_result_path_that_contain_scenario_data

    def get_list_of_all_relevant_folders_or_files(self, result_path: str, folder_or_filename: str) -> List[str]:
        """Get a list of all folders or files which you want to analyze."""

        # choose which path to check
        path_to_check = os.path.join(result_path, "**", folder_or_filename)

        list_of_paths_first_order = list(glob.glob(path_to_check))

        # if in these paths no result data folder can be found check in subfolders for it
        path_to_check = os.path.join(result_path, "**", "**", folder_or_filename)  # type: ignore
        list_of_paths_second_order = list(glob.glob(path_to_check))

        path_to_check = os.path.join(result_path, "**", "**", "**", folder_or_filename)  # type: ignore
        list_of_paths_third_order = list(glob.glob(path_to_check))

        list_with_all_paths_to_check = (
            list_of_paths_first_order + list_of_paths_second_order + list_of_paths_third_order
        )

        return list_with_all_paths_to_check

    def import_data_from_file(self, paths_to_check: List[str], analyze_yearly_or_hourly_data: str) -> List:
        """Import data from result files."""
        log.information("Importing result_data_for_scenario_evaluation from csv files.")

        all_csv_files = []

        if analyze_yearly_or_hourly_data == ResultDataTypeEnum.HOURLY.name:
            kind_of_data_set = "hourly"
        elif analyze_yearly_or_hourly_data == ResultDataTypeEnum.YEARLY.name:
            kind_of_data_set = "yearly"
        elif analyze_yearly_or_hourly_data == ResultDataTypeEnum.DAILY.name:
            kind_of_data_set = "daily"
        elif analyze_yearly_or_hourly_data == ResultDataTypeEnum.MONTHLY.name:
            kind_of_data_set = "monthly"
        else:
            raise ValueError("analyze_yearly_or_hourly_data was not found in the datacollectorenum class.")

        for folder in paths_to_check:  # type: ignore
            for file in os.listdir(folder):  # type: ignore
                # get yearly or hourly data
                if kind_of_data_set in file and file.endswith(".csv"):
                    all_csv_files.append(os.path.join(folder, file))  # type: ignore

        return all_csv_files

    def make_dictionaries_with_simulation_duration_keys(
        self, simulation_duration_to_check: str, all_csv_files: List[str],
    ) -> Dict:
        """Make dictionaries containing csv files of hourly or yearly data and according to the simulation duration of the data."""

        dict_of_csv_data: Dict[str, Any] = {}

        dict_of_csv_data[f"{simulation_duration_to_check}"] = []

        # open file config and check if they have wanted simulation duration
        for file in all_csv_files:
            parent_folder = os.path.abspath(os.path.join(file, os.pardir))  # type: ignore
            for file1 in os.listdir(parent_folder):
                if ".json" in file1:
                    with open(os.path.join(parent_folder, file1), "r", encoding="utf-8") as openfile:
                        json_file = json.load(openfile)
                        simulation_duration = json_file["scenarioDataInformation"].get("duration in days")
                        if int(simulation_duration_to_check) == int(simulation_duration):
                            dict_of_csv_data[f"{simulation_duration}"].append(file)
                        else:
                            raise ValueError(
                                f"The simulation_duration_to_check of {simulation_duration_to_check} is different,"
                                f"to the simulation duration of {simulation_duration} found in the scenario data information json in the result folders."
                            )

        # raise error if dict is empty
        if bool(dict_of_csv_data) is False:
            raise ValueError(
                "The dictionary is empty. Maybe no data was collected. Please check your parameters again."
            )

        return dict_of_csv_data

    def alternative_read_csv_and_generate_pandas_dataframe(
        self,
        dict_of_csv_to_read: Dict[str, list[str]],
        time_resolution_of_data_set: Any,
        rename_scenario: bool = False,
        parameter_key: Optional[str] = None,
        list_with_parameter_key_values: Optional[List[Any]] = None,
        list_with_module_config_dicts: Optional[List[Any]] = None,
    ) -> str:
        """Read the csv files and generate the result dataframe."""
        log.information(f"Read csv files and generate result dataframes for {time_resolution_of_data_set}.")

        if not dict_of_csv_to_read:
            raise ValueError("The input dictionary dict_of_csv_to_read is empty.")

        simulation_duration_key = list(dict_of_csv_to_read.keys())[0]
        csv_data_list = dict_of_csv_to_read[simulation_duration_key]

        if not csv_data_list:
            raise ValueError("csv_data_list is empty.")

        # Initialize dictionaries to hold data
        dict_with_all_data: Dict[str, defaultdict] = {
            "Index": defaultdict(list),
            "Input": defaultdict(list),
            "Output": defaultdict(list),
        }

        for house_index, csv_file in enumerate(csv_data_list):
            log.information(f"Reading data from house number {house_index}")
            dataframe = pd.read_csv(csv_file)
            set_of_variables = dataframe["variable"].unique()

            for variable in set_of_variables:
                filtered_df = dataframe[dataframe["variable"] == variable]

                if time_resolution_of_data_set != ResultDataTypeEnum.YEARLY.name and "time" in dataframe.columns:
                    # Process time series data
                    time_values = filtered_df["time"]
                    filtered_df = filtered_df[~time_values.str.isalnum()]
                    # if in timeseries data there are also yearly data contained (like KPIs) which cause an empty filtered_df, the variable should be skipped
                    if filtered_df.empty:
                        continue
                    for time_value, value in zip(filtered_df["time"], filtered_df["value"]):
                        dict_with_all_data["Output"][time_value].append(value)
                elif time_resolution_of_data_set == ResultDataTypeEnum.YEARLY.name:
                    # Process yearly data
                    dict_with_all_data["Output"][str(dataframe["year"].iloc[0])].append(filtered_df["value"].iloc[0])

                original_scenario_name = str(filtered_df["scenario"].iloc[0])
                if not original_scenario_name:
                    raise ValueError(
                        "The scenario variable of the current dataframe is empty. Please set a scenario name for your simulations."
                    )

                try:
                    hash_number = re.findall(r"-?\d+", original_scenario_name)[-1]
                except IndexError:
                    hash_number = 1
                    rename_scenario = False

                # Add input and metadata values to the dictionary
                dict_with_all_data["Input"]["model"].append(filtered_df["model"].iloc[0])
                dict_with_all_data["Input"]["region"].append(filtered_df["region"].iloc[0])
                dict_with_all_data["Input"]["hash"].append(hash_number)

                # Add outputs to dict
                dict_with_all_data["Output"]["variable"].append(variable)
                dict_with_all_data["Output"]["unit"].append(filtered_df["unit"].tolist()[0])

                if list_with_module_config_dicts is not None:
                    module_config_dict = list_with_module_config_dicts[house_index]
                    for key, value in module_config_dict.items():
                        dict_with_all_data["Input"][key].append(value)

                dict_with_all_data["Index"]["Index"].append(house_index)

                if rename_scenario and parameter_key and list_with_parameter_key_values:
                    value = list_with_parameter_key_values[house_index]
                    final_scenario_name = f"{value if isinstance(value, str) else round(value, 1)}"
                else:
                    final_scenario_name = original_scenario_name

                dict_with_all_data["Input"]["scenario"].append(final_scenario_name)

        if not dict_with_all_data:
            raise ValueError("The dict_with_all_data is empty")

        # create multiindex columns
        multi_index_columns = pd.MultiIndex.from_tuples(
            [(key1, key2) for key1, v1_dict in dict_with_all_data.items() for key2 in v1_dict.keys()],
            names=["first", "second"],
        )
        # add everything to the dataframe
        appended_dataframe = pd.DataFrame(
            {
                (key1, key2): value_list2
                for key1, v1_dict in dict_with_all_data.items()
                for key2, value_list2 in v1_dict.items()
            },
            columns=multi_index_columns,
        )

        if self.data_format_type == DataFormatEnum.CSV.name:
            # create filename
            filename = self.store_scenario_data_with_the_right_name_and_in_the_right_path(
                result_data_folder=self.result_data_folder,
                simulation_duration_key=simulation_duration_key,
                time_resolution_of_data_set=time_resolution_of_data_set,
                parameter_key=parameter_key,
                data_format_type="csv",
                scenario_analysis_config_name=self.scenario_analysis_config_name,
            )
            # save file compressed
            # appended_dataframe.to_csv(f"{filename}.gz", compression="gzip")
            appended_dataframe.to_csv(filename)

        elif self.data_format_type == DataFormatEnum.XLSX.name:
            # create filename
            filename = self.store_scenario_data_with_the_right_name_and_in_the_right_path(
                result_data_folder=self.result_data_folder,
                simulation_duration_key=simulation_duration_key,
                time_resolution_of_data_set=time_resolution_of_data_set,
                parameter_key=parameter_key,
                data_format_type="xlsx",
                scenario_analysis_config_name=self.scenario_analysis_config_name,
            )
            # save file (use zip64 for handling large excel files)
            with pd.ExcelWriter(filename, engine="xlsxwriter") as writer:  # pylint: disable=abstract-class-instantiated
                workbook = writer.book
                workbook.use_zip64()
                appended_dataframe.to_excel(writer, sheet_name="Sheet1")

        else:
            raise ValueError(f"Only data format types xlsx or csv are implemented. Here it is {self.data_format_type}.")

        log.information(f"Saving result dataframe here: {filename}")

        del appended_dataframe
        del dict_with_all_data
        return filename

    def store_scenario_data_with_the_right_name_and_in_the_right_path(
        self,
        result_data_folder: str,
        simulation_duration_key: str,
        time_resolution_of_data_set: str,
        data_format_type: str,
        scenario_analysis_config_name: str,
        parameter_key: Optional[str] = None,
    ) -> str:
        """Store csv files in the result data folder with the right filename and path."""

        if time_resolution_of_data_set == ResultDataTypeEnum.HOURLY.name:
            kind_of_data_set = "hourly"
        elif time_resolution_of_data_set == ResultDataTypeEnum.YEARLY.name:
            kind_of_data_set = "yearly"
        elif time_resolution_of_data_set == ResultDataTypeEnum.DAILY.name:
            kind_of_data_set = "daily"
        elif time_resolution_of_data_set == ResultDataTypeEnum.MONTHLY.name:
            kind_of_data_set = "monthly"
        else:
            raise ValueError("This kind of data was not found in the datacollectorenum class.")

        if parameter_key is not None:
            path_for_file = os.path.join(
                result_data_folder, f"data_different_{parameter_key}s", f"{simulation_duration_key}_days",
            )
        else:
            path_for_file = os.path.join(result_data_folder, "data_all_parameters", f"{simulation_duration_key}_days",)
        if os.path.exists(path_for_file) is False:
            os.makedirs(path_for_file)

        filename = os.path.join(path_for_file, f"result_df_{kind_of_data_set}_{scenario_analysis_config_name}.{data_format_type}",)
        return filename

    def get_default_config(self, path_to_default_config: Optional[str]) -> Any:
        """Get default config."""

        if path_to_default_config is not None and ".json" in path_to_default_config:
            with open(path_to_default_config, "r", encoding="utf-8") as openfile:  # type: ignore
                default_config_dict = json.load(openfile)

        else:
            raise ValueError("The default config is not in .json format.")

        return default_config_dict

    def read_scenario_data_json_config_and_compare_to_default_config(
        self,
        default_config_dict: Dict[str, Any],
        path_to_scenario_data_folder: str,
        list_with_csv_files: List[Any],
        list_with_parameter_key_values: List[Any],
        list_with_module_configs: List[Any],
        parameter_key: str,
    ) -> tuple[List[Any], List[Any], List[Any]]:
        """Read json config in result_data_for_scenario_evaluation folder and compare with default config."""

        for file in os.listdir(path_to_scenario_data_folder):
            if ".json" in file:
                with open(os.path.join(path_to_scenario_data_folder, file), "r", encoding="utf-8") as openfile:  # type: ignore
                    config_dict = json.load(openfile)
                    my_module_config_dict = config_dict["myModuleConfig"]
                    scenario_name = config_dict["systemName"]

                    # for paper: reference scenario without use of pv should have a share of maximum pv power of 0
                    if "ref_" in scenario_name:
                        try:
                            my_module_config_dict["share_of_maximum_pv_power"] = 0
                        except Exception as ecx:
                            raise KeyError(
                                "The key share of maximum pv power does not exist in the module dict. Unable this function if it not needed."
                            ) from ecx

        # check if module config and default config have any keys in common
        if len(set(default_config_dict).intersection(my_module_config_dict)) == 0:
            raise KeyError(
                f"The module config of the folder {path_to_scenario_data_folder} should contain the keys of the default config,",
                "otherwise their values cannot be compared.",
            )
        # check if there is a module config which is equal to default config

        if all(item in my_module_config_dict.items() for item in default_config_dict.items()):
            self.path_of_scenario_data_executed_with_default_config = path_to_scenario_data_folder

        # for each parameter different than the default config parameter, get the respective path to the folder
        # and also create a dict with the parameter, value pairs

        list_with_csv_files.append(path_to_scenario_data_folder)
        list_with_parameter_key_values.append(my_module_config_dict[parameter_key])

        list_with_module_configs.append(my_module_config_dict)

        # add to each item in the dict also the default system setup if the default system setup exists

        if self.path_of_scenario_data_executed_with_default_config != "":
            list_with_csv_files.append(self.path_of_scenario_data_executed_with_default_config)
            list_with_parameter_key_values.append(default_config_dict[parameter_key])

            list_with_module_configs.append(default_config_dict)

        return (
            list_with_csv_files,
            list_with_parameter_key_values,
            list_with_module_configs,
        )

    def read_module_config_if_exist_and_write_in_dataframe(
        self,
        default_config_dict: Dict[str, Any],
        path_to_scenario_data_folder: str,
        list_with_module_configs: List[Any],
        list_with_csv_files: List[Any],
    ) -> Tuple[List, List]:
        """Read module config if possible and write to dataframe."""

        for file in os.listdir(path_to_scenario_data_folder):
            if ".json" in file:
                with open(os.path.join(path_to_scenario_data_folder, file), "r", encoding="utf-8") as openfile:  # type: ignore
                    config_dict = json.load(openfile)
                    my_module_config_dict = config_dict["myModuleConfig"]

        # check if module config and default config have any keys in common
        if len(set(default_config_dict).intersection(my_module_config_dict)) == 0:
            raise KeyError(
                f"The module config of the folder {path_to_scenario_data_folder} should contain the keys of the default config,",
                "otherwise their values cannot be compared.",
            )

        list_with_module_configs.append(my_module_config_dict)
        list_with_csv_files.append(path_to_scenario_data_folder)

        return (
            list_with_module_configs,
            list_with_csv_files,
        )

    def go_through_all_result_data_folders_and_collect_file_paths_according_to_parameters(
        self,
        list_with_result_data_folders: List[str],
        default_config_dict: Dict[str, Any],
        parameter_key: Optional[str],
    ) -> tuple[List[Any], List[Any], List[Any]]:
        """Order result files according to different parameters."""

        list_with_module_configs: List = []
        list_with_csv_files: List = []
        list_with_parameter_key_values: List = []

        for folder in list_with_result_data_folders:  # type: ignore
            if parameter_key is None:
                (
                    list_with_module_configs,
                    list_with_csv_files,
                ) = self.read_module_config_if_exist_and_write_in_dataframe(
                    default_config_dict=default_config_dict,
                    path_to_scenario_data_folder=folder,
                    list_with_module_configs=list_with_module_configs,
                    list_with_csv_files=list_with_csv_files,
                )
                list_with_parameter_key_values = []

            else:
                (
                    list_with_csv_files,
                    list_with_parameter_key_values,
                    list_with_module_configs,
                ) = self.read_scenario_data_json_config_and_compare_to_default_config(
                    default_config_dict=default_config_dict,
                    path_to_scenario_data_folder=folder,
                    list_with_csv_files=list_with_csv_files,
                    list_with_parameter_key_values=list_with_parameter_key_values,
                    list_with_module_configs=list_with_module_configs,
                    parameter_key=parameter_key,
                )

        return (
            list_with_csv_files,
            list_with_parameter_key_values,
            list_with_module_configs,
        )

    def check_for_duplicates_in_dict(self, dictionary_to_check: Dict[str, Any], key: str) -> List:
        """Check for duplicates and return index of where the duplicates are found."""

        indices_of_duplicates = [
            index for index, value in enumerate(dictionary_to_check[key]) if value in dictionary_to_check[key][:index]
        ]

        return indices_of_duplicates

    def go_through_all_scenario_data_folders_and_check_if_module_configs_are_double_somewhere(
        self, list_of_result_folder_paths_to_check: List[str]
    ) -> List[Any]:
        """Go through all result folders and remove the system_setups that are duplicated."""
        if list_of_result_folder_paths_to_check == []:
            raise ValueError(
                "No HiSim results could be found in the results folder. Please check if you are collecting results from the correct folder."
            )

        list_of_all_module_configs = []
        list_of_result_folders_which_have_only_unique_configs = []
        for folder in list_of_result_folder_paths_to_check:
            for file in os.listdir(folder):
                if ".json" in file:
                    filename = os.path.join(folder, file)
                    with open(filename, "r", encoding="utf-8") as openfile:  # type: ignore
                        config_dict = json.load(openfile)
                        try:
                            my_module_config_dict = config_dict["myModuleConfig"]
                            my_module_config_dict.update(
                                {"duration in days": config_dict["scenarioDataInformation"].get("duration in days")}
                            )
                            my_module_config_dict.update({"model": config_dict["scenarioDataInformation"].get("model")})
                            my_module_config_dict.update(
                                {"model": config_dict["scenarioDataInformation"].get("scenario")}
                            )
                        except Exception as exc:
                            raise KeyError(
                                f"The file {filename} does not contain any key called myModuleConfig."
                            ) from exc
                        # prevent to add modules with same module config and same simulation duration twice
                        if my_module_config_dict not in list_of_all_module_configs:
                            list_of_all_module_configs.append(my_module_config_dict)
                            list_of_result_folders_which_have_only_unique_configs.append(os.path.join(folder))

            # get folders with duplicates
            list_with_duplicates = []
            if folder not in list_of_result_folders_which_have_only_unique_configs:
                whole_parent_folder = os.path.abspath(os.path.join(folder, os.pardir))
                list_with_duplicates.append(whole_parent_folder)

        print(
            f"The following folders seem to be duplicated: {list_with_duplicates}. Number: {len(list_with_duplicates)}."
        )
        # if list is not empty
        if list_with_duplicates:
            answer = input("Do you want to delete the duplicated folders?")
            if answer.upper() in ["Y", "YES"]:
                for folder in list_with_duplicates:
                    shutil.rmtree(folder, ignore_errors=True)
                print("All folders with duplicated results are deleted.")
            elif answer.upper() in ["N", "NO"]:
                print("These folders won't be deleted.")
            else:
                print("The answer must be yes or no.")

        return list_of_result_folders_which_have_only_unique_configs



# clean

""" Contains functions for initializing and connecting components.

The functions are all called in modular_household.
"""

import json
from os import path
from typing import Any, List, Optional, Tuple

import pandas as pd
from utspclient.helpers.lpgpythonbindings import JsonReference

import hisim.loadtypes as lt
from hisim import utils
from hisim.component import Component
from hisim.components import (
    advanced_battery_bslib,
    advanced_ev_battery_bslib,
    building,
    controller_l1_building_heating,
    controller_l1_generic_ev_charge,
    controller_l1_heatpump,
    controller_l1_chp,
    controller_l1_electrolyzer,
    controller_l2_energy_management_system,
    generic_car,
    generic_chp,
    generic_electrolyzer,
    generic_heat_pump_modular,
    generic_heat_source,
    generic_hot_water_storage_modular,
    generic_hydrogen_storage,
    generic_pv_system,
    generic_smart_device,
    weather,
    loadprofilegenerator_utsp_connector
)

from hisim.components.configuration import HouseholdWarmWaterDemandConfig
from hisim.simulator import SimulationParameters


def get_heating_system_efficiency(
    heating_system_installed: lt.HeatingSystems, water_vs_heating: lt.InandOutputType
) -> float:
    """Reads in type of heating system and returns related efficiency values.

    :param heating_system_installed: type of installed heating system
    :type heating_system_installed: lt.HeatingSystems
    :param water_vs_heating: Heating vs. WaterHeating
    :type water_vs_heating: lt.InandOutputType
    :return: efficiency of the selected heater
    :rtype: float
    """

    efficiency_data = pd.read_csv(utils.HISIMPATH["heater_efficiencies"], encoding="utf-8", index_col=0).astype("float")

    return float(efficiency_data.loc[heating_system_installed.value][water_vs_heating.value])


def configure_pv_system(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_weather: weather.Weather,
    production: List,
    pv_peak_power: Optional[float],
    count: int,
) -> Tuple[List, int]:
    """Sets PV System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_weather: Weather
        The initialized Weather component.
    production: List
        List of Components with Parameter Production.
    pv_peak_power: float or None
        The peak power of the PV panel. In case of None default is used.
    count: int
        Integer tracking component hierachy for EMS.

    """
    if pv_peak_power is not None:
        my_pv_system_config = generic_pv_system.PVSystem.get_default_config(
            power_in_watt=pv_peak_power,
            source_weight=count,
        )
    else:
        my_pv_system_config = generic_pv_system.PVSystem.get_default_config(
            source_weight=count,
        )
    my_pv_system_config.location = my_weather.weather_config.location
    count += 1
    my_pv_system = generic_pv_system.PVSystem(
        my_simulation_parameters=my_simulation_parameters,
        config=my_pv_system_config,
    )
    my_pv_system.connect_only_predefined_connections(my_weather)
    my_sim.add_component(my_pv_system)
    production.append(my_pv_system)

    return production, count


def configure_smart_devices(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    count: int,
    smart_devices_included: bool,
) -> Tuple[List[generic_smart_device.SmartDevice], int]:
    """Sets smart devices without controllers.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    smart_devices_included: bool
        True if smart devices (washing machine, dish washer, etc.) are actually smart or surplus controlled.
    count: int
        Integer tracking component hierachy for EMS.

    """
    filepath = path.join(utils.HISIMPATH["utsp_reports"], "FlexibilityEvents.HH1.json")
    device_collection = []
    with open(filepath, mode="r", encoding="utf-8") as jsonfile:
        strfile = json.load(jsonfile)

    for elem in strfile:
        if elem["Device"]["Name"] in device_collection:
            pass
        else:
            device_collection.append(elem["Device"]["Name"])

    # create all smart devices
    my_smart_devices: List[generic_smart_device.SmartDevice] = []
    for device in device_collection:
        my_smart_devices.append(
            generic_smart_device.SmartDevice(
                config=generic_smart_device.SmartDeviceConfig(
                    name="SmartDevice",
                    building_name="BUI1",
                    identifier=device,
                    source_weight=count,
                    smart_devices_included=smart_devices_included,
                ),
                my_simulation_parameters=my_simulation_parameters,
            )
        )
        my_sim.add_component(my_smart_devices[-1])
        count += 1

    return my_smart_devices, count


def configure_cars(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    count: int,
    ev_included: bool,
    my_occupancy_instance: loadprofilegenerator_utsp_connector.UtspLpgConnector,
) -> Tuple[List[generic_car.Car], int]:
    """Sets smart devices without controllers.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    count: int
        Integer tracking component hierachy for EMS.
    ev_included: bool
        True if Car is electric, False if it is diesel.
    my_occupancy_instance: loadprofilegenerator_utsp_connector.UtspLpgConnector,
        Instance of load profile generator giving the necessary mobility data (mobility is related!)

    """
    # get data of all available cars
    my_car_information = generic_car.GenericCarInformation(my_occupancy_instance=my_occupancy_instance)

    # create all cars
    my_cars: List[generic_car.Car] = []
    for car_information_dict in my_car_information.data_dict_for_car_component.values():
        # decide if they are diesel driven or electricity driven and initialize config
        if ev_included:
            my_car_config = generic_car.CarConfig.get_default_ev_config()
        else:
            my_car_config = generic_car.CarConfig.get_default_diesel_config()
        # reset name and source weight
        my_car_config.source_weight = count
        my_cars.append(
            generic_car.Car(
                my_simulation_parameters=my_simulation_parameters,
                config=my_car_config,
                data_dict_with_car_information=car_information_dict,
            )
        )
        my_sim.add_component(my_cars[-1])
        count += 1

    return my_cars, count


def configure_ev_batteries(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_cars: List[generic_car.Car],
    charging_station_set: Optional[JsonReference],
    mobility_set: JsonReference,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    controllable: bool,
) -> List:
    """Sets batteries and controllers of electric vehicles.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_cars: List[Car]
        List of initilized cars.
    charging_station_set: ChargingStationSets
        Encoding of the charging station.
    mobility_set: TransportationDeviceSets
        Encoding of the available cars.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    controllable: bool
        True if battery of electric vehicle is charged with surplus.

    """
    ev_capacities = []

    if mobility_set.Name is None:
        raise Exception("For EV configuration mobility set is obligatory.")

    if charging_station_set is not None:
        charging_power = float((charging_station_set.Name or "").split("with ")[1].split(" kW")[0])
    else:
        raise Exception("For EV configuration charging station set is obligatory.")

    for car in my_cars:
        car_battery_config = advanced_ev_battery_bslib.CarBatteryConfig.get_default_config()
        car_battery_config.source_weight = car.config.source_weight
        car_battery_config.p_inv_custom = charging_power * 1e3
        my_carbattery = advanced_ev_battery_bslib.CarBattery(
            my_simulation_parameters=my_simulation_parameters, config=car_battery_config
        )
        ev_capacities.append(car_battery_config.e_bat_custom)

        car_battery_controller_config = controller_l1_generic_ev_charge.ChargingStationConfig.get_default_config(
            charging_station_set=charging_station_set
        )
        car_battery_controller_config.source_weight = car.config.source_weight
        car_battery_controller_config.lower_threshold_charging_power = (
            charging_power * 1e3 * 0.1
        )  # 10 % of charging power for acceptable efficiencies
        if controllable:
            car_battery_controller_config.battery_set = 0.4  # lower threshold for soc of car battery in clever case
        my_controller_carbattery = controller_l1_generic_ev_charge.L1Controller(
            my_simulation_parameters=my_simulation_parameters,
            config=car_battery_controller_config,
        )
        my_controller_carbattery.connect_only_predefined_connections(car)
        my_controller_carbattery.connect_only_predefined_connections(my_carbattery)
        my_carbattery.connect_only_predefined_connections(my_controller_carbattery)
        my_sim.add_component(my_carbattery)
        my_sim.add_component(my_controller_carbattery)

        if controllable:
            my_electricity_controller.add_component_input_and_connect(
                source_object_name=my_carbattery.component_name,
                source_component_output=my_carbattery.AcBatteryPower,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[
                    lt.ComponentType.CAR_BATTERY,
                    lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
                ],
                source_weight=my_carbattery.source_weight,
            )

            electricity_target = my_electricity_controller.add_component_output(
                source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
                source_tags=[
                    lt.ComponentType.CAR_BATTERY,
                    lt.InandOutputType.ELECTRICITY_TARGET,
                ],
                source_weight=my_controller_carbattery.source_weight,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                output_description="Target Electricity for EV Battery Controller. ",
            )

            my_controller_carbattery.connect_dynamic_input(
                input_fieldname=controller_l1_generic_ev_charge.L1Controller.ElectricityTarget,
                src_object=electricity_target,
            )

    return ev_capacities


def configure_smart_controller_for_smart_devices(
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    my_smart_devices: List[generic_smart_device.SmartDevice],
) -> None:
    """Sets l3 controller for smart devices.

    Parameters
    ----------
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    my_smart_devices: List[SmartDevice]
        List of initilized smart devices.

    """

    for elem in my_smart_devices:
        my_electricity_controller.add_component_input_and_connect(
            source_object_name=elem.component_name,
            source_component_output=elem.ElectricityOutput,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            source_tags=[
                lt.ComponentType.SMART_DEVICE,
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
            ],
            source_weight=elem.source_weight,
        )

        electricity_to_smart_device = my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[
                lt.ComponentType.SMART_DEVICE,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=elem.source_weight,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for Smart Device Controller. ",
        )

        elem.connect_dynamic_input(
            input_fieldname=generic_smart_device.SmartDevice.ElectricityTarget,
            src_object=electricity_to_smart_device,
        )


def configure_battery(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    battery_capacity: Optional[float],
    count: int,
) -> int:
    """Sets advanced battery system with surplus controller.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    battery_capacity: float or None
        Capacity of the battery in Wh. In case of None default is used.
    count: int
        Integer tracking component hierachy for EMS.

    """
    if battery_capacity is not None:
        my_advanced_battery_config = advanced_battery_bslib.BatteryConfig(
            custom_battery_capacity_generic_in_kilowatt_hour=battery_capacity,
            custom_pv_inverter_power_generic_in_watt=battery_capacity * 0.5 * 1e3,
            source_weight=count,
            system_id="SG1",
            building_name="BUI1",
            name="Battery",
            charge_in_kwh=0,
            discharge_in_kwh=0,
            co2_footprint=battery_capacity * 130.7,
            cost=battery_capacity * 535.81,
            lifetime=10,  # todo set correct values
            lifetime_in_cycles=5e3,  # todo set correct values
            maintenance_cost_as_percentage_of_investment=0.02,
        )
    else:
        my_advanced_battery_config = advanced_battery_bslib.BatteryConfig.get_default_config()
        my_advanced_battery_config.source_weight = count
    count += 1
    my_advanced_battery = advanced_battery_bslib.Battery(
        my_simulation_parameters=my_simulation_parameters,
        config=my_advanced_battery_config,
    )

    my_electricity_controller.add_component_input_and_connect(
        source_object_name=my_advanced_battery.component_name,
        source_component_output=my_advanced_battery.AcBatteryPowerUsed,
        source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT,
        source_tags=[lt.ComponentType.BATTERY, lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED],
        source_weight=my_advanced_battery.source_weight,
    )

    electricity_to_or_from_battery_target = my_electricity_controller.add_component_output(
        source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
        source_tags=[
            lt.ComponentType.BATTERY,
            lt.InandOutputType.ELECTRICITY_TARGET,
        ],
        source_weight=my_advanced_battery.source_weight,
        source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT,
        output_description="Target electricity for Battery Control. ",
    )

    my_advanced_battery.connect_dynamic_input(
        input_fieldname=advanced_battery_bslib.Battery.LoadingPowerInput,
        src_object=electricity_to_or_from_battery_target,
    )
    my_sim.add_component(my_advanced_battery)

    return count


def configure_water_heating(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_occupancy: loadprofilegenerator_utsp_connector.UtspLpgConnector,
    water_heating_system_installed: lt.HeatingSystems,
    count: int,
    number_of_apartments: float,
) -> Tuple[generic_hot_water_storage_modular.HotWaterStorage, int]:
    """Sets Boiler with Heater, L1 Controller and L2 Controller for Water Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_occupancy: Occupancy
        The initialized occupancy component.
    water_heating_system_installed: str
        Type of installed WaterHeatingSystem
    count: int
        Integer tracking component hierachy for EMS.
    number_of_apartments: float
        from building component

    """
    fuel_translator = {
        lt.HeatingSystems.GAS_HEATING: lt.LoadTypes.GAS,
        lt.HeatingSystems.OIL_HEATING: lt.LoadTypes.OIL,
        lt.HeatingSystems.DISTRICT_HEATING: lt.LoadTypes.DISTRICTHEATING,
    }
    heater_config = generic_heat_source.HeatSourceConfig.get_default_config_waterheating_with_district_heating()
    heater_config.fuel = fuel_translator[water_heating_system_installed]
    heater_config.efficiency = get_heating_system_efficiency(
        heating_system_installed=water_heating_system_installed, water_vs_heating=lt.InandOutputType.HEATING
    )
    heater_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_dhw(
        "DHW" + water_heating_system_installed.value
    )
    [heater_config.source_weight, heater_l1_config.source_weight] = [count] * 2
    count += 1
    boiler_config = (
        generic_hot_water_storage_modular.StorageConfig.get_scaled_config_for_boiler_to_number_of_apartments(
            number_of_apartments=number_of_apartments
        )
    )
    boiler_config.compute_default_cycle(
        temperature_difference_in_kelvin=heater_l1_config.t_max_heating_in_celsius
        - heater_l1_config.t_min_heating_in_celsius
    )

    heater_config.power_th = (
        my_occupancy.max_hot_water_demand
        * (4180 / 3600)
        * 0.5
        * (3600 / my_simulation_parameters.seconds_per_timestep)
        * (
            HouseholdWarmWaterDemandConfig.ww_temperature_demand
            - HouseholdWarmWaterDemandConfig.temperature_difference_hot
            - HouseholdWarmWaterDemandConfig.freshwater_temperature
        )
        * my_occupancy.scaling_factor_according_to_number_of_apartments
    )

    my_boiler = generic_hot_water_storage_modular.HotWaterStorage(
        my_simulation_parameters=my_simulation_parameters, config=boiler_config
    )
    my_boiler.connect_only_predefined_connections(my_occupancy)
    my_sim.add_component(my_boiler)

    my_heater_controller_l1 = controller_l1_heatpump.L1HeatPumpController(
        my_simulation_parameters=my_simulation_parameters, config=heater_l1_config
    )
    my_heater_controller_l1.connect_only_predefined_connections(my_boiler)
    my_sim.add_component(my_heater_controller_l1)

    my_heater = generic_heat_source.HeatSource(config=heater_config, my_simulation_parameters=my_simulation_parameters)
    my_heater.connect_only_predefined_connections(my_heater_controller_l1)

    my_sim.add_component(my_heater)

    my_boiler.connect_only_predefined_connections(my_heater)
    return my_boiler, count


def configure_water_heating_electric(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_occupancy: loadprofilegenerator_utsp_connector.UtspLpgConnector,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    my_weather: weather.Weather,
    water_heating_system_installed: lt.HeatingSystems,
    controllable: bool,
    count: int,
    number_of_apartments: float,
) -> Tuple[generic_hot_water_storage_modular.HotWaterStorage, int]:
    """Sets Boiler with Heater, L1 Controller and L2 Controller for Water Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_occupancy: Occupancy
        The initialized occupancy component.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    my_weather: Weather
        The initialized Weather component.
    water_heating_system_installed: str
        Type of installed WaterHeatingSystem
    controllable: bool
        True if control of heating device is smart, False if not.
    count: int
        Integer tracking component hierachy for EMS.
    number_of_apartments: float
        from building component

    """
    if water_heating_system_installed == lt.HeatingSystems.HEAT_PUMP:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_waterheating()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_dhw(
            "DHWHeatPumpController"
        )
    elif water_heating_system_installed == lt.HeatingSystems.ELECTRIC_HEATING:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_waterheating_electric()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_dhw(
            "BoilerHeatingController"
        )

    [heatpump_config.source_weight, heatpump_l1_config.source_weight] = [count] * 2
    count += 1

    heatpump_config.power_th = (
        my_occupancy.max_hot_water_demand
        * (4180 / 3600)
        * 0.5
        * (3600 / my_simulation_parameters.seconds_per_timestep)
        * (
            HouseholdWarmWaterDemandConfig.ww_temperature_demand
            - HouseholdWarmWaterDemandConfig.temperature_difference_hot
            - HouseholdWarmWaterDemandConfig.freshwater_temperature
        )
        * my_occupancy.scaling_factor_according_to_number_of_apartments
    )
    boiler_config = (
        generic_hot_water_storage_modular.StorageConfig.get_scaled_config_for_boiler_to_number_of_apartments(
            number_of_apartments=number_of_apartments
        )
    )
    boiler_config.compute_default_cycle(
        temperature_difference_in_kelvin=heatpump_l1_config.t_max_heating_in_celsius
        - heatpump_l1_config.t_min_heating_in_celsius
    )

    my_boiler = generic_hot_water_storage_modular.HotWaterStorage(
        my_simulation_parameters=my_simulation_parameters, config=boiler_config
    )
    my_boiler.connect_only_predefined_connections(my_occupancy)
    my_sim.add_component(my_boiler)

    my_heatpump_controller_l1 = controller_l1_heatpump.L1HeatPumpController(
        my_simulation_parameters=my_simulation_parameters, config=heatpump_l1_config
    )
    my_heatpump_controller_l1.connect_only_predefined_connections(my_boiler)
    my_sim.add_component(my_heatpump_controller_l1)

    my_heatpump = generic_heat_pump_modular.ModularHeatPump(
        config=heatpump_config, my_simulation_parameters=my_simulation_parameters
    )
    my_heatpump.connect_only_predefined_connections(my_weather)
    my_heatpump.connect_only_predefined_connections(my_heatpump_controller_l1)
    my_sim.add_component(my_heatpump)
    my_boiler.connect_only_predefined_connections(my_heatpump)

    if controllable:
        my_heatpump_controller_l1.connect_input(
            my_heatpump_controller_l1.StorageTemperatureModifier,
            my_electricity_controller.component_name,
            my_electricity_controller.DomesticHotWaterStorageTemperatureModifier,
        )
        my_electricity_controller.add_component_input_and_connect(
            source_object_name=my_heatpump.component_name,
            source_component_output=my_heatpump.ElectricityOutput,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            source_tags=[
                lt.ComponentType.HEAT_PUMP,
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
            ],
            source_weight=my_heatpump.config.source_weight,
        )

        my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[
                lt.ComponentType.HEAT_PUMP,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=my_heatpump.config.source_weight,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for heat pump.",
        )

    else:
        my_electricity_controller.add_component_input_and_connect(
            source_object_name=my_heatpump.component_name,
            source_component_output=my_heatpump.ElectricityOutput,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            source_tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
            source_weight=999,
        )
    return my_boiler, count


def configure_heating(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_building: building.Building,
    heating_system_installed: lt.HeatingSystems,
    heating_season: List[int],
    count: int,
) -> Tuple[Component, int]:
    """Sets Heater, L1 Controller and L2 Controller for Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    heating_system_installed: str
        Type of installed HeatingSystem
    heating_season: List[int]
        Contains first and last day of heating season.
    count: int
        Integer tracking component hierachy for EMS.

    """
    fuel_translator = {
        lt.HeatingSystems.GAS_HEATING: lt.LoadTypes.GAS,
        lt.HeatingSystems.OIL_HEATING: lt.LoadTypes.OIL,
        lt.HeatingSystems.DISTRICT_HEATING: lt.LoadTypes.DISTRICTHEATING,
    }
    heater_config = generic_heat_source.HeatSourceConfig.get_default_config_heating()
    heater_config.fuel = fuel_translator[heating_system_installed]
    heater_config.efficiency = get_heating_system_efficiency(
        heating_system_installed=heating_system_installed, water_vs_heating=lt.InandOutputType.HEATING
    )
    heater_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller(
        heating_system_installed.value
    )

    # set power of heating system according to maximal power demand
    heater_config.power_th = my_building.my_building_information.max_thermal_building_demand_in_watt
    heater_l1_config.day_of_heating_season_end = heating_season[0]
    heater_l1_config.day_of_heating_season_begin = heating_season[1]

    [heater_config.source_weight, heater_l1_config.source_weight] = [count] * 2
    count += 1

    my_heater_controller_l1 = controller_l1_heatpump.L1HeatPumpController(
        my_simulation_parameters=my_simulation_parameters, config=heater_l1_config
    )
    # connect building manually
    my_heater_controller_l1.connect_input(
        input_fieldname=my_heater_controller_l1.StorageTemperature,
        src_object_name=my_building.component_name,
        src_field_name=my_building.TemperatureMeanThermalMass,
    )
    my_sim.add_component(my_heater_controller_l1)

    my_heater = generic_heat_source.HeatSource(config=heater_config, my_simulation_parameters=my_simulation_parameters)
    my_heater.connect_only_predefined_connections(my_heater_controller_l1)
    my_sim.add_component(my_heater)

    my_building.connect_input(
        input_fieldname=my_building.ThermalPowerDelivered,
        src_object_name=my_heater.component_name,
        src_field_name=my_heater.ThermalPowerDelivered,
    )

    return my_heater, count


def configure_heating_electric(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_building: building.Building,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    my_weather: weather.Weather,
    heating_system_installed: lt.HeatingSystems,
    heatpump_power: float,
    controllable: bool,
    heating_season: List[int],
    count: int,
) -> Tuple[Component, int]:
    """Sets Heater, L1 Controller and L2 Controller for Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    my_weather: Weather
        The initialized Weather component.
    heating_system_installed: str
        Type of installed HeatingSystem
    heatpump_power: float,
        Power of heat pump in multiples of default.
    controllable: bool
        True if control of heating device is smart, False if not.
    heating_season: List[int]
        Contains first and last day of heating season.
    count: int
        Integer tracking component hierachy for EMS.

    """
    if heating_system_installed == lt.HeatingSystems.HEAT_PUMP:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_heating()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller(
            "HeatigHeatPumpController"
        )
    elif heating_system_installed == lt.HeatingSystems.ELECTRIC_HEATING:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_heating_electric()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller(
            "ElectricHeatingController"
        )

    heatpump_config.power_th = my_building.my_building_information.max_thermal_building_demand_in_watt * heatpump_power
    heatpump_l1_config.day_of_heating_season_end = heating_season[0]
    heatpump_l1_config.day_of_heating_season_begin = heating_season[1]
    [heatpump_config.source_weight, heatpump_l1_config.source_weight] = [count] * 2
    count += 1

    my_heatpump_controller_l1 = controller_l1_heatpump.L1HeatPumpController(
        my_simulation_parameters=my_simulation_parameters, config=heatpump_l1_config
    )
    my_heatpump_controller_l1.connect_only_predefined_connections(my_building)
    my_sim.add_component(my_heatpump_controller_l1)

    my_heatpump = generic_heat_pump_modular.ModularHeatPump(
        config=heatpump_config, my_simulation_parameters=my_simulation_parameters
    )
    my_heatpump.connect_only_predefined_connections(my_weather)
    my_heatpump.connect_only_predefined_connections(my_heatpump_controller_l1)
    my_sim.add_component(my_heatpump)

    if controllable:
        my_heatpump_controller_l1.connect_input(
            my_heatpump_controller_l1.StorageTemperatureModifier,
            my_electricity_controller.component_name,
            my_electricity_controller.BuildingIndoorTemperatureModifier,
        )
        my_electricity_controller.add_component_input_and_connect(
            source_object_name=my_heatpump.component_name,
            source_component_output=my_heatpump.ElectricityOutput,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            source_tags=[
                lt.ComponentType.HEAT_PUMP,
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
            ],
            source_weight=my_heatpump.config.source_weight,
        )

        my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[
                lt.ComponentType.HEAT_PUMP,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=my_heatpump.config.source_weight,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for HeatingHeat Pump. ",
        )
    else:
        my_electricity_controller.add_component_input_and_connect(
            source_object_name=my_heatpump.component_name,
            source_component_output=my_heatpump.ElectricityOutput,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            source_tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
            source_weight=999,
        )

    my_building.connect_input(
        input_fieldname=my_building.ThermalPowerDelivered,
        src_object_name=my_heatpump.component_name,
        src_field_name=my_heatpump.ThermalPowerDelivered,
    )

    return my_heatpump, count


def configure_heating_with_buffer_electric(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_building: building.Building,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    my_weather: weather.Weather,
    heating_system_installed: lt.HeatingSystems,
    heatpump_power: float,
    buffer_volume: float,
    controllable: bool,
    heating_season: List[int],
    count: int,
) -> Tuple:
    """Sets Heater, L1 Controller and L2 Controller for Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    my_electricity_controller: L2GenericEnergyManagementSystem
        The initialized electricity controller.
    my_weather: Weather
        The initialized Weather component.
    heating_system_installed: str
        Type of installed HeatingSystem.
    heatpump_power: float
        Power of heat pump in multiples of default.
    buffer_volume: float
        Volume of buffer storage in multiples of default.
    controllable: bool
        True if control of heating device is smart, False if not.
    heating_season: List[int]
        Contains first and last day of heating season.
    count: int
        Integer tracking component hierachy for EMS.

    """
    if heating_system_installed == lt.HeatingSystems.HEAT_PUMP:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_heating()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_buffer(
            "BufferHeatPumpController"
        )
    elif heating_system_installed == lt.HeatingSystems.ELECTRIC_HEATING:
        heatpump_config = generic_heat_pump_modular.HeatPumpConfig.get_default_config_heating_electric()
        heatpump_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_buffer(
            "BufferElectricHeatingController"
        )

    heatpump_config.power_th = my_building.my_building_information.max_thermal_building_demand_in_watt * heatpump_power
    heatpump_l1_config.day_of_heating_season_end = heating_season[0] + 1
    heatpump_l1_config.day_of_heating_season_begin = heating_season[1] - 1
    [heatpump_config.source_weight, heatpump_l1_config.source_weight] = [count] * 2
    count += 1

    buffer_config = generic_hot_water_storage_modular.StorageConfig.get_default_config_buffer(
        power=float(my_building.my_building_information.max_thermal_building_demand_in_watt)
    )
    buffer_config.compute_default_volume(
        time_in_seconds=heatpump_l1_config.min_idle_time_in_seconds,
        temperature_difference_in_kelvin=heatpump_l1_config.t_max_heating_in_celsius
        - heatpump_l1_config.t_min_heating_in_celsius,
        multiplier=buffer_volume,
    )
    buffer_config.compute_default_cycle(
        temperature_difference_in_kelvin=heatpump_l1_config.t_max_heating_in_celsius
        - heatpump_l1_config.t_min_heating_in_celsius
    )

    building_heating_controller_config = (
        controller_l1_building_heating.L1BuildingHeatingConfig.get_default_config_heating("buffer")
    )
    building_heating_controller_config.day_of_heating_season_end = heating_season[0]
    building_heating_controller_config.day_of_heating_season_begin = heating_season[1]
    building_heating_controller_config.t_buffer_activation_threshold_in_celsius = (
        heatpump_l1_config.t_max_heating_in_celsius
    )
    [buffer_config.source_weight, building_heating_controller_config.source_weight] = [count] * 2
    count += 1

    my_buffer = generic_hot_water_storage_modular.HotWaterStorage(
        my_simulation_parameters=my_simulation_parameters, config=buffer_config
    )
    my_sim.add_component(my_buffer)

    my_heatpump_controller_l1 = controller_l1_heatpump.L1HeatPumpController(
        my_simulation_parameters=my_simulation_parameters, config=heatpump_l1_config
    )
    my_heatpump_controller_l1.connect_only_predefined_connections(my_buffer)
    my_sim.add_component(my_heatpump_controller_l1)
    my_heatpump = generic_heat_pump_modular.ModularHeatPump(
        config=heatpump_config, my_simulation_parameters=my_simulation_parameters
    )
    my_heatpump.connect_only_predefined_connections(my_weather)
    my_heatpump.connect_only_predefined_connections(my_heatpump_controller_l1)
    my_sim.add_component(my_heatpump)

    my_buffer_controller = controller_l1_building_heating.L1BuildingHeatController(
        my_simulation_parameters=my_simulation_parameters,
        config=building_heating_controller_config,
    )
    my_buffer_controller.connect_only_predefined_connections(my_building)
    my_buffer_controller.connect_only_predefined_connections(my_buffer)
    my_sim.add_component(my_buffer_controller)

    if controllable:
        my_heatpump_controller_l1.connect_input(
            my_heatpump_controller_l1.StorageTemperatureModifier,
            my_electricity_controller.component_name,
            my_electricity_controller.DomesticHotWaterStorageTemperatureModifier,
        )
        my_buffer_controller.connect_input(
            my_buffer_controller.BuildingTemperatureModifier,
            my_electricity_controller.component_name,
            my_electricity_controller.BuildingIndoorTemperatureModifier,
        )

        my_electricity_controller.add_component_input_and_connect(
            source_object_name=my_heatpump.component_name,
            source_component_output=my_heatpump.ElectricityOutput,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            source_tags=[
                lt.ComponentType.HEAT_PUMP,
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
            ],
            source_weight=my_heatpump.config.source_weight,
        )

        my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[
                lt.ComponentType.HEAT_PUMP,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=my_heatpump.config.source_weight,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for HeatingHeat Pump. ",
        )

    else:
        my_electricity_controller.add_component_input_and_connect(
            source_object_name=my_heatpump.component_name,
            source_component_output=my_heatpump.ElectricityOutput,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            source_tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
            source_weight=999,
        )

    my_buffer_controller.connect_only_predefined_connections(my_electricity_controller)
    my_buffer.connect_only_predefined_connections(my_buffer_controller)
    my_buffer.connect_only_predefined_connections(my_heatpump)
    my_building.connect_input(
        input_fieldname=my_building.ThermalPowerDelivered,
        src_object_name=my_buffer.component_name,
        src_field_name=my_buffer.PowerFromHotWaterStorage,
    )

    return my_heatpump, my_buffer, count


def configure_heating_with_buffer(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_building: building.Building,
    heating_system_installed: lt.HeatingSystems,
    buffer_volume: float,
    heating_season: List[int],
    count: int,
) -> Tuple:
    """Sets Heater, L1 Controller and L2 Controller for Heating System.

    Parameters
    ----------
    my_sim: str
        filename of orginal built system setup.
    my_simulation_parameters: SimulationParameters
        The simulation parameters.
    my_building: Building
        The initialized building component.
    my_weather: Weather
        The initialized Weather component.
    heating_system_installed: str
        Type of installed HeatingSystem.
    buffer_volume: float
        Volume of buffer storage in multiples of default. In case of None one is used.
    heating_season: List[int]
        Contains first and last day of heating season.
    count: int
        Integer tracking component hierachy for EMS.

    """
    fuel_translator = {
        lt.HeatingSystems.GAS_HEATING: lt.LoadTypes.GAS,
        lt.HeatingSystems.OIL_HEATING: lt.LoadTypes.OIL,
        lt.HeatingSystems.DISTRICT_HEATING: lt.LoadTypes.DISTRICTHEATING,
    }
    heater_config = generic_heat_source.HeatSourceConfig.get_default_config_heating()
    heater_config.fuel = fuel_translator[heating_system_installed]
    heater_config.power_th = my_building.my_building_information.max_thermal_building_demand_in_watt
    heater_config.efficiency = get_heating_system_efficiency(
        heating_system_installed=heating_system_installed, water_vs_heating=lt.InandOutputType.HEATING
    )
    heater_l1_config = controller_l1_heatpump.L1HeatPumpConfig.get_default_config_heat_source_controller_buffer(
        "Buffer" + heating_system_installed.value + "Controller"
    )

    heater_l1_config.day_of_heating_season_end = heating_season[0] + 1
    heater_l1_config.day_of_heating_season_begin = heating_season[1] - 1
    [heater_config.source_weight, heater_l1_config.source_weight] = [count] * 2
    count += 1

    buffer_config = generic_hot_water_storage_modular.StorageConfig.get_default_config_buffer(
        power=float(my_building.my_building_information.max_thermal_building_demand_in_watt)
    )
    buffer_config.compute_default_volume(
        time_in_seconds=heater_l1_config.min_idle_time_in_seconds,
        temperature_difference_in_kelvin=heater_l1_config.t_max_heating_in_celsius
        - heater_l1_config.t_min_heating_in_celsius,
        multiplier=buffer_volume,
    )
    buffer_config.compute_default_cycle(
        temperature_difference_in_kelvin=heater_l1_config.t_max_heating_in_celsius
        - heater_l1_config.t_min_heating_in_celsius
    )

    building_heating_controller_config = (
        controller_l1_building_heating.L1BuildingHeatingConfig.get_default_config_heating("buffer")
    )
    building_heating_controller_config.day_of_heating_season_end = heating_season[0]
    building_heating_controller_config.day_of_heating_season_begin = heating_season[1] - 1
    building_heating_controller_config.t_buffer_activation_threshold_in_celsius = (
        heater_l1_config.t_max_heating_in_celsius
    )
    [buffer_config.source_weight, building_heating_controller_config.source_weight] = [count] * 2
    count += 1

    my_buffer = generic_hot_water_storage_modular.HotWaterStorage(
        my_simulation_parameters=my_simulation_parameters, config=buffer_config
    )
    my_sim.add_component(my_buffer)
    my_heater_controller_l1 = controller_l1_heatpump.L1HeatPumpController(
        my_simulation_parameters=my_simulation_parameters, config=heater_l1_config
    )
    my_heater_controller_l1.connect_only_predefined_connections(my_buffer)
    my_sim.add_component(my_heater_controller_l1)

    my_heater = generic_heat_source.HeatSource(config=heater_config, my_simulation_parameters=my_simulation_parameters)
    my_heater.connect_only_predefined_connections(my_heater_controller_l1)
    my_sim.add_component(my_heater)

    my_buffer_controller = controller_l1_building_heating.L1BuildingHeatController(
        my_simulation_parameters=my_simulation_parameters,
        config=building_heating_controller_config,
    )
    my_buffer_controller.connect_only_predefined_connections(my_building)
    my_sim.add_component(my_buffer_controller)
    my_buffer.connect_only_predefined_connections(my_buffer_controller)
    my_buffer.connect_only_predefined_connections(my_heater)
    my_building.connect_input(
        input_fieldname=my_building.ThermalPowerDelivered,
        src_object_name=my_buffer.component_name,
        src_field_name=my_buffer.PowerFromHotWaterStorage,
    )

    return my_heater, my_buffer, count


def configure_chp(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_building: building.Building,
    my_boiler: generic_hot_water_storage_modular.HotWaterStorage,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    chp_power: float,
    controllable: bool,
    count: int,
) -> int:
    """Sets up natural gas CHP. It heats the DHW storage and the building in winter.

    :param my_sim: Simulation class.
    :type my_sim: Any
    :param my_simulation_parameters: Simulation parameters for HiSIM calculation.
    :type my_simulation_parameters: SimulationParameters
    :param my_building: Building of the HiSIM system setup.
    :type my_building: building.Building
    :param my_boiler: Hot water storage of the HiSIM system setup.
    :type my_boiler: generic_hot_water_storage_modular.HotWaterStorage
    :param my_electricity_controller: Energy Management System of the HiSIM system setup
    :type my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem
    :param chp_power: Power of the CHP in multiples of default (<=1).
    :type chp_power: float
    :param controllable: When True, surplus control of Energy Management System is activated.
    :type controllable: bool
    :param count: Number of component outputs relevant in the energy management system.
    :type count: int
    :return: New counter variable (+1).
    :rtype: int
    """

    # configure and add chp controller
    chp_controller_config = controller_l1_chp.L1CHPControllerConfig.get_default_config_chp()
    chp_controller_config.source_weight = count

    # size chp power to hot water storage size
    my_boiler.config.compute_default_cycle(
        temperature_difference_in_kelvin=chp_controller_config.t_max_dhw_in_celsius
        - chp_controller_config.t_min_dhw_in_celsius
    )
    chp_power = (
        chp_power
        * (my_boiler.config.energy_full_cycle or 1)
        * 3.6e6
        / chp_controller_config.min_operation_time_in_seconds
        or 1
    )

    # configure and add chp
    chp_config = generic_chp.CHPConfig.get_default_config_chp(thermal_power=chp_power)
    chp_config.source_weight = count
    my_chp = generic_chp.SimpleCHP(my_simulation_parameters=my_simulation_parameters, config=chp_config)

    # add treshold electricity to chp controller and add it to simulation
    chp_controller_config.electricity_threshold = chp_config.p_el / 2
    my_chp_controller = controller_l1_chp.L1CHPController(
        my_simulation_parameters=my_simulation_parameters, config=chp_controller_config
    )
    my_chp_controller.connect_only_predefined_connections(my_boiler)
    my_chp_controller.connect_only_predefined_connections(my_building)
    my_sim.add_component(my_chp_controller)

    # connect chp with controller intputs and add it to simulation
    my_chp.connect_only_predefined_connections(my_chp_controller)
    my_sim.add_component(my_chp)

    # connect thermal power output of CHP
    my_boiler.connect_only_predefined_connections(my_chp)
    my_building.connect_input(
        input_fieldname=my_building.ThermalPowerCHP,
        src_object_name=my_chp.component_name,
        src_field_name=my_chp.ThermalPowerOutputBuilding,
    )

    my_electricity_controller.add_component_input_and_connect(
        source_object_name=my_chp.component_name,
        source_component_output="ElectricityOutput",
        source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT,
        source_tags=[lt.ComponentType.CHP, lt.InandOutputType.ELECTRICITY_PRODUCTION],
        source_weight=my_chp.config.source_weight,
    )

    # connect to EMS electricity controller
    if controllable:
        ems_target_electricity = my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[
                lt.ComponentType.CHP,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=my_chp.config.source_weight,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for CHP. ",
        )

        my_chp_controller.connect_dynamic_input(
            input_fieldname=my_chp_controller.ElectricityTarget,
            src_object=ems_target_electricity,
        )

    # counting variable
    count += 1

    return count


def configure_chp_with_buffer(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_buffer: generic_hot_water_storage_modular.HotWaterStorage,
    my_boiler: generic_hot_water_storage_modular.HotWaterStorage,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    chp_power: float,
    controllable: bool,
    count: int,
) -> int:
    """Sets up natural gas CHP. It heats the DHW storage and the buffer storage for heating.

    :param my_sim: Simulation class.
    :type my_sim: Any
    :param my_simulation_parameters: Simulation parameters for HiSIM calculation.
    :type my_simulation_parameters: SimulationParameters
    :param my_buffer: Buffer storage for heating of the HISIM system setup
    :type my_buffer: generic_hot_water_storage_modular.HotWaterStorage
    :param my_boiler: Hot water storage of the HiSIM system setup.
    :type my_boiler: generic_hot_water_storage_modular.HotWaterStorage
    :param my_electricity_controller: Energy Management System of the HiSIM system setup
    :type my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem
    :param chp_power: Power of the CHP in multiples of default (<=1)
    :type chp_power: float
    :param controllable: When True, surplus control of Energy Management System is activated.
    :type controllable: bool
    :param count: Number of component outputs relevant in the energy management system.
    :type count: int
    :return: New counter variable (+1).
    :rtype: int
    """

    # configure chp controller
    chp_controller_config = controller_l1_chp.L1CHPControllerConfig.get_default_config_chp_with_buffer()
    chp_controller_config.source_weight = count

    # size chp power to hot water storage size
    my_boiler.config.compute_default_cycle(
        temperature_difference_in_kelvin=chp_controller_config.t_max_dhw_in_celsius
        - chp_controller_config.t_min_dhw_in_celsius
    )
    chp_power = (
        chp_power
        * (my_boiler.config.energy_full_cycle or 1)
        * 3.6e6
        / chp_controller_config.min_operation_time_in_seconds
    )

    # configure and add chp
    chp_config = generic_chp.CHPConfig.get_default_config_chp(thermal_power=chp_power)
    chp_config.source_weight = count
    my_chp = generic_chp.SimpleCHP(
        my_simulation_parameters=my_simulation_parameters,
        config=chp_config,
    )

    # add chop controller and adopt electricity threshold
    chp_controller_config.electricity_threshold = chp_config.p_el / 2
    my_chp_controller = controller_l1_chp.L1CHPController(
        my_simulation_parameters=my_simulation_parameters,
        config=chp_controller_config,
    )
    my_chp_controller.connect_only_predefined_connections(my_boiler)
    my_chp_controller.connect_input(
        input_fieldname=my_chp_controller.BuildingTemperature,
        src_object_name=my_buffer.component_name,
        src_field_name=my_buffer.TemperatureMean,
    )
    my_sim.add_component(my_chp_controller)

    # connect chp with controller intputs and add it to simulation
    my_chp.connect_only_predefined_connections(my_chp_controller)
    my_sim.add_component(my_chp)

    # connect power output of CHP
    my_boiler.connect_only_predefined_connections(my_chp)
    my_buffer.connect_input(
        input_fieldname=my_buffer.ThermalPowerCHP,
        src_object_name=my_chp.component_name,
        src_field_name=my_chp.ThermalPowerOutputBuilding,
    )

    my_electricity_controller.add_component_input_and_connect(
        source_object_name=my_chp.component_name,
        source_component_output="ElectricityOutput",
        source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT,
        source_tags=[lt.ComponentType.CHP, lt.InandOutputType.ELECTRICITY_PRODUCTION],
        source_weight=my_chp.config.source_weight,
    )

    # connect to EMS electricity controller
    if controllable:
        ems_target_electricity = my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[
                lt.ComponentType.CHP,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=my_chp.config.source_weight,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for CHP. ",
        )

        my_chp_controller.connect_dynamic_input(
            input_fieldname=my_chp_controller.ElectricityTarget,
            src_object=ems_target_electricity,
        )

    # counting variable
    count += 1

    return count


def configure_electrolyzer_and_h2_storage(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_chp: generic_chp.SimpleCHP,
    my_chp_controller: controller_l1_chp.L1CHPController,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    electrolyzer_power: float,
    h2_storage_size: float,
    fuel_cell_power: float,
    count: int,
) -> int:
    """Configures electrolyzer and h2 storage with fuel cell already defined.

    (in configure_elctrolysis_h2storage_fuelcell_system (_with_buffer))

    :param my_sim: Simulation class.
    :type my_sim: Any
    :param my_simulation_parameters: Simulation parameters for HiSIM calculation.
    :type my_simulation_parameters: SimulationParameters
    :param my_chp: Fuel cell component of the HiSIM system setup
    :type my_chp: generic_chp.CHP
    :param my_chp_controller: Fuel cell controller component of the HiSIM system setup
    :type my_chp_controller: controller_l1_chp.L1CHPController
    :param my_electricity_controller: Energy management system component of the HiSIM system setup
    :type my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem
    :param electrolyzer_power: Power of the electrolyzer in Watt
    :type electrolyzer_power: float
    :param h2_storage_size: Size of the hydrogen storage in capacity for storing kg of hydrogen
    :type h2_storage_size: float
    :param fuel_cell_power: power of the configured fuel cell
    :type fuel_cell_power: float
    :param count: Number of component outputs relevant in the energy management system.
    :type count: int
    :return: New counter variable (+1).
    :rtype: int
    """

    # electrolyzer default configuration
    electrolyzer_config = generic_electrolyzer.GenericElectrolyzerConfig.get_default_config(p_el=electrolyzer_power)
    electrolyzer_config.source_weight = count

    # electrolyzer controller default configuration and counting variable
    electrolyzer_controller_config = controller_l1_electrolyzer.L1ElectrolyzerControllerConfig.get_default_config()
    electrolyzer_controller_config.source_weight = count
    electrolyzer_controller_config.p_min_electrolyzer = electrolyzer_config.min_power
    count += 1

    # electrolyzer
    my_electrolyzer = generic_electrolyzer.GenericElectrolyzer(
        my_simulation_parameters=my_simulation_parameters, config=electrolyzer_config
    )
    my_sim.add_component(my_electrolyzer)

    # run time controller of electrolyzer
    my_electrolyzer_controller = controller_l1_electrolyzer.L1GenericElectrolyzerController(
        my_simulation_parameters=my_simulation_parameters,
        config=electrolyzer_controller_config,
    )
    my_sim.add_component(my_electrolyzer_controller)
    my_electrolyzer.connect_only_predefined_connections(my_electrolyzer_controller)
    print(my_electrolyzer)

    # electricity controller of fuel cell
    my_electricity_controller.add_component_input_and_connect(
        source_object_name=my_electrolyzer.component_name,
        source_component_output=my_electrolyzer.ElectricityOutput,
        source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT,
        source_tags=[
            lt.ComponentType.ELECTROLYZER,
            lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
        ],
        source_weight=my_electrolyzer.config.source_weight,
    )
    electricity_to_electrolyzer_target = my_electricity_controller.add_component_output(
        source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
        source_tags=[
            lt.ComponentType.ELECTROLYZER,
            lt.InandOutputType.ELECTRICITY_TARGET,
        ],
        source_weight=my_electrolyzer.config.source_weight,
        source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT,
        output_description="Target electricity for electrolyzer. ",
    )
    my_electrolyzer_controller.connect_dynamic_input(
        input_fieldname=controller_l1_electrolyzer.L1GenericElectrolyzerController.ElectricityTarget,
        src_object=electricity_to_electrolyzer_target,
    )

    # hydrogen storage default configuration
    h2_storage_config = generic_hydrogen_storage.GenericHydrogenStorageConfig.get_default_config(
        capacity=h2_storage_size,
        max_charging_rate=electrolyzer_power / (3.6e3 * 3.939e4),
        max_discharging_rate=fuel_cell_power / (3.6e3 * 3.939e4),
        source_weight=count,
    )
    my_h2storage = generic_hydrogen_storage.GenericHydrogenStorage(
        my_simulation_parameters=my_simulation_parameters, config=h2_storage_config
    )
    my_h2storage.connect_only_predefined_connections(my_electrolyzer)
    my_h2storage.connect_only_predefined_connections(my_chp)
    my_sim.add_component(my_h2storage)

    my_electrolyzer_controller.connect_only_predefined_connections(my_h2storage)
    my_chp_controller.connect_only_predefined_connections(my_h2storage)

    return count


def configure_elctrolysis_h2storage_fuelcell_system(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_building: building.Building,
    my_boiler: generic_hot_water_storage_modular.HotWaterStorage,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    fuel_cell_power: float,
    h2_storage_size: float,
    electrolyzer_power: float,
    controllable: bool,
    count: int,
) -> int:
    """Sets electrolysis, H2-storage and chp system.

    :param my_sim: Simulation class.
    :type my_sim: Any
    :param my_simulation_parameters: Simulation parameters for HiSIM calculation.
    :type my_simulation_parameters: SimulationParameters
    :param my_building: Building component of the HiSIM system setup.
    :type my_building: building.Building
    :param my_boiler: Hot water storage (for drain hot water) component of the HiSIM system setup.
    :type my_boiler: generic_hot_water_storage_modular.HotWaterStorage
    :param my_electricity_controller:Energy Management System controller component of the HiSIM system_setups.
    :type my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem
    :param fuel_cell_power: Power of the fuel cell in Watt
    :type fuel_cell_power: float
    :param h2_storage_size: Size of the hydrogen storage in capacity for storing kg of hydrogen
    :type h2_storage_size: float
    :param electrolyzer_power: Power of the electrolyzer in Watt.
    :type electrolyzer_power: float
    :param controllable: Electricity based control of the fuel cell considered or not.
    :type controllable: bool
    :param count: Number of component outputs relevant in the energy management system.
    :type count: int
    """

    # configure and add chp controller
    chp_controller_config = controller_l1_chp.L1CHPControllerConfig.get_default_config_fuel_cell()
    chp_controller_config.source_weight = count

    # size chp power to hot water storage size
    my_boiler.config.compute_default_cycle(
        temperature_difference_in_kelvin=chp_controller_config.t_max_dhw_in_celsius
        - chp_controller_config.t_min_dhw_in_celsius
    )
    fuel_cell_power = (
        fuel_cell_power
        * (my_boiler.config.energy_full_cycle or 1)
        * 3.6e6
        / chp_controller_config.min_operation_time_in_seconds
        or 1
    )

    # configure and add chp
    chp_config = generic_chp.CHPConfig.get_default_config_fuelcell(thermal_power=fuel_cell_power)
    chp_config.source_weight = count
    my_chp = generic_chp.SimpleCHP(my_simulation_parameters=my_simulation_parameters, config=chp_config)

    # add treshold electricity to chp controller and add it to simulation
    chp_controller_config.electricity_threshold = chp_config.p_el / 2
    my_chp_controller = controller_l1_chp.L1CHPController(
        my_simulation_parameters=my_simulation_parameters, config=chp_controller_config
    )
    my_chp_controller.connect_only_predefined_connections(my_boiler)
    my_chp_controller.connect_only_predefined_connections(my_building)
    my_sim.add_component(my_chp_controller)

    # connect chp with controller intputs and add it to simulation
    my_chp.connect_only_predefined_connections(my_chp_controller)
    my_sim.add_component(my_chp)

    # connect thermal power output of CHP
    my_boiler.connect_only_predefined_connections(my_chp)
    my_building.connect_input(
        input_fieldname=my_building.ThermalPowerCHP,
        src_object_name=my_chp.component_name,
        src_field_name=my_chp.ThermalPowerOutputBuilding,
    )

    my_electricity_controller.add_component_input_and_connect(
        source_object_name=my_chp.component_name,
        source_component_output="ElectricityOutput",
        source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT,
        source_tags=[lt.ComponentType.CHP, lt.InandOutputType.ELECTRICITY_PRODUCTION],
        source_weight=my_chp.config.source_weight,
    )

    # connect to EMS electricity controller
    if controllable:
        ems_target_electricity = my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[
                lt.ComponentType.CHP,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=my_chp.config.source_weight,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for CHP. ",
        )

        my_chp_controller.connect_dynamic_input(
            input_fieldname=my_chp_controller.ElectricityTarget,
            src_object=ems_target_electricity,
        )

    # counting variable
    count += 1

    count = configure_electrolyzer_and_h2_storage(
        my_sim=my_sim,
        my_simulation_parameters=my_simulation_parameters,
        my_chp=my_chp,
        my_chp_controller=my_chp_controller,
        my_electricity_controller=my_electricity_controller,
        electrolyzer_power=electrolyzer_power,
        h2_storage_size=h2_storage_size,
        fuel_cell_power=fuel_cell_power,
        count=count,
    )

    return count


def configure_elctrolysis_h2storage_fuelcell_system_with_buffer(
    my_sim: Any,
    my_simulation_parameters: SimulationParameters,
    my_buffer: generic_hot_water_storage_modular.HotWaterStorage,
    my_boiler: generic_hot_water_storage_modular.HotWaterStorage,
    my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem,
    fuel_cell_power: float,
    h2_storage_size: float,
    electrolyzer_power: float,
    controllable: bool,
    count: int,
) -> int:
    """Sets electrolysis, H2-storage and chp system.

    :param my_sim: Simulation class.
    :type my_sim: Any
    :param my_simulation_parameters: Simulation parameters for HiSIM calculation.
    :type my_simulation_parameters: SimulationParameters
    :param my_buffer: Buffer storage component of the HiSIM system setup
    :type my_buffer: generic_hot_water_storage_modular.HotWaterStorage
    :param my_boiler: Hot water storage (for drain hot water) component of the HiSIM system setup.
    :type my_boiler: generic_hot_water_storage_modular.HotWaterStorage
    :param my_electricity_controller:Energy Management System controller component of the HiSIM system_setups.
    :type my_electricity_controller: controller_l2_energy_management_system.L2GenericEnergyManagementSystem
    :param fuel_cell_power: Power of the fuel cell in Watt
    :type fuel_cell_power: float
    :param h2_storage_size: Size of the hydrogen storage in capacity for storing kg of hydrogen
    :type h2_storage_size: float
    :param electrolyzer_power: Power of the electrolyzer in Watt.
    :type electrolyzer_power: float
    :param controllable: Electricity based control of the fuel cell considered or not.
    :type controllable: bool
    :param count: Number of component outputs relevant in the energy management system.
    :type count: int
    """

    # configure and add chp controller
    chp_controller_config = controller_l1_chp.L1CHPControllerConfig.get_default_config_fuel_cell()
    chp_controller_config.source_weight = count

    # size chp power to hot water storage size
    my_boiler.config.compute_default_cycle(
        temperature_difference_in_kelvin=chp_controller_config.t_max_dhw_in_celsius
        - chp_controller_config.t_min_dhw_in_celsius
    )
    fuel_cell_power = (
        fuel_cell_power
        * (my_boiler.config.energy_full_cycle or 1)
        * 3.6e6
        / chp_controller_config.min_operation_time_in_seconds
        or 1
    )

    # configure and add chp
    chp_config = generic_chp.CHPConfig.get_default_config_fuelcell(thermal_power=fuel_cell_power)
    chp_config.source_weight = count
    my_chp = generic_chp.SimpleCHP(my_simulation_parameters=my_simulation_parameters, config=chp_config)

    # add treshold electricity to chp controller and add it to simulation
    chp_controller_config.electricity_threshold = chp_config.p_el / 2
    my_chp_controller = controller_l1_chp.L1CHPController(
        my_simulation_parameters=my_simulation_parameters, config=chp_controller_config
    )
    my_chp_controller.connect_only_predefined_connections(my_boiler)
    my_chp_controller.connect_input(
        input_fieldname=my_chp_controller.BuildingTemperature,
        src_object_name=my_buffer.component_name,
        src_field_name=my_buffer.TemperatureMean,
    )
    my_sim.add_component(my_chp_controller)

    # connect chp with controller intputs and add it to simulation
    my_chp.connect_only_predefined_connections(my_chp_controller)
    my_sim.add_component(my_chp)

    # connect thermal power output of CHP
    my_boiler.connect_only_predefined_connections(my_chp)
    my_buffer.connect_input(
        input_fieldname=my_buffer.ThermalPowerCHP,
        src_object_name=my_chp.component_name,
        src_field_name=my_chp.ThermalPowerOutputBuilding,
    )

    my_electricity_controller.add_component_input_and_connect(
        source_object_name=my_chp.component_name,
        source_component_output="ElectricityOutput",
        source_load_type=lt.LoadTypes.ELECTRICITY,
        source_unit=lt.Units.WATT,
        source_tags=[lt.ComponentType.CHP, lt.InandOutputType.ELECTRICITY_PRODUCTION],
        source_weight=my_chp.config.source_weight,
    )

    # connect to EMS electricity controller
    if controllable:
        ems_target_electricity = my_electricity_controller.add_component_output(
            source_output_name=lt.InandOutputType.ELECTRICITY_TARGET,
            source_tags=[
                lt.ComponentType.CHP,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=my_chp.config.source_weight,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for CHP. ",
        )

        my_chp_controller.connect_dynamic_input(
            input_fieldname=my_chp_controller.ElectricityTarget,
            src_object=ems_target_electricity,
        )

    # counting variable
    count += 1

    count = configure_electrolyzer_and_h2_storage(
        my_sim=my_sim,
        my_simulation_parameters=my_simulation_parameters,
        my_chp=my_chp,
        my_chp_controller=my_chp_controller,
        my_electricity_controller=my_electricity_controller,
        electrolyzer_power=electrolyzer_power,
        h2_storage_size=h2_storage_size,
        fuel_cell_power=fuel_cell_power,
        count=count,
    )

    return count
